"""
Course data structure and related functions.
"""
import re
from typing import Optional, Tuple, List, Iterator
from functools import partial
from concurrent.futures import ThreadPoolExecutor
from yaml import dump
from lxml.etree import Element
from .fetch import fetch
from .parse import COMBINATOR

COURSE_PATTERN = r"([a-zA-Z]+)[ ]?(\d{3})"


class Course:
    def __init__(self, subject: str, number: int, hours: Optional[int], label: Optional[str],
                 description: Optional[str], schedule_info: Optional[str],
                 prereqs: List[frozenset], coreqs: List[frozenset], constraints: List[str],
                 raw: Element):
        """
        :param subject: the name of the course's subject (like 'CS')
        :param number: the course number (like 421)
        :param hours: the number of credit hours the course is
        :param label: the course label (like 'Programming Languages & Compilers')
        :param description: the course description
        :param schedule_info: info about course scheduling
        :param prereqs: prerequisite requirements
        :param coreqs: corequisite requirements
        :param constraints: additional constraints on taking the course
        :param raw: a ``lxml.etree.Element`` object with the raw XML course data
        """
        self.subject = subject
        self.number = number
        self.name = "{0} {1}".format(self.subject, self.number)

        self.hours = hours
        self.label = label
        self.description = description
        self.schedule_info = schedule_info

        self.prereqs = prereqs
        self.coreqs = coreqs
        self.constraints = constraints

        self.raw = raw

    def serialize(self) -> str:
        """
        Generates YAML representation of course.

        :return: YAML dict with the course's name as the single key and its attributes as the value
        """
        attrs = {k: v for k, v in self.__dict__.items() if k not in ("name", "raw")}
        return dump({self.name: attrs}, sort_keys=False)

    def __repr__(self) -> str:
        return repr(self.__dict__)


def get_course_name(course_name: str) -> Tuple[str, str]:
    """
    Helper method to validate a course name string.

    :param course_name: valid course name like 'CS 225'
    :return: subject, course number
    """
    if not isinstance(course_name, str):
        raise TypeError("non-string passed as course name")
    m = re.match(COURSE_PATTERN, course_name)
    if m:
        subject = m.group(1).upper()
        number = m.group(2)
        return subject, number
    else:
        raise ValueError("{} is not a valid course name".format(course_name))


def get_course(course_name: str, year: Optional[str] = None,
               quarter: Optional[str] = None, redirect: bool = False) -> Course:
    """
    Returns parsed ``Course`` object.

    :param year: year to get catalog from
    :param quarter: one of ``{'fall', 'spring', 'winter', 'summer'}``
                    (any case, defaults to spring)
    :param course_name: valid course name (like 'CS 225', 'cs225', etc.)
    :param redirect: if ``True``, courses that are the same will redirect to the root course
                     (for instance, MATH 213 will simply return the ``Course`` object for CS 173)
    :return: a ``Course`` object
    """
    subject, number = get_course_name(course_name)
    element = fetch(subject=subject, number=number, year=year, quarter=quarter)
    description = element.find("description")
    if description is not None:
        description = description.text
        if redirect:
            redirect_match = re.search("[Ss]ee ({})".format(COURSE_PATTERN), description)
            if redirect_match:
                return get_course(redirect_match.group(1), redirect=False)

    parsed = COMBINATOR(element)
    return Course(subject=subject, number=number, description=description, raw=element, **parsed)


def get_courses(course_name_iterator: Iterator[str], year: Optional[str] = None,
                quarter: Optional[str] = None, max_workers: Optional[int] = None) -> List[Course]:
    """
    Wrapper function around get_course that uses ``concurrent.futures.ThreadPoolExecutor`` to
    fetch data concurrently.

    :param course_name_iterator: iterator of course names
    :param year: year to get catalog from
    :param quarter: one of ``{'fall', 'spring', 'winter', 'summer'}``
                    (any case, defaults to spring)
    :param max_workers: number of workers to use (defaults to ``os.cpu_count() + 4``)
    :return: list of ``Course`` objects
    """
    with ThreadPoolExecutor(max_workers) as e:
        return e.map(partial(get_course, year=year, quarter=quarter), course_name_iterator)



"""
Parses Elements generated from calls to fetch() into dicts
"""
import re
from typing import Callable, Dict, List, Optional, Tuple, Iterable
from lxml.etree import Element
from os import path
from lark import Lark
from .utils import reduce_tree

with open(path.join(path.dirname(__file__), "prerequisites.lark")) as f:
    GRAMMAR = f.read()
PREREQUISITE_PARSER = Lark(GRAMMAR)


class Parser:
    """
    Callable which associates a parse function with named parameters.
    """

    def __init__(self, params: Iterable[str], parse_func: Callable):
        self.params = params
        self.parse_func = parse_func

    def __call__(self, element: Element) -> Tuple:
        """
        Wrapper for ``self.parse_func(element)``.

        if output is not tuple, converts to tuple (for easy zipping with ``self.params``)
        :param element: ``Element`` object
        :return: Tuple with value(s) corresponding to self.params
        """
        val = self.parse_func(element)
        if isinstance(val, tuple):
            return val
        else:
            return (val,)


class ParserCombinator:
    """
    Higher order callable which takes parsers as inputs and feeds an ``Element`` into all of them.
    """

    def __init__(self, *parsers: Parser):
        """
        :param parsers: ``Parser`` objects created through ``generate()``
        """
        self.parsers = parsers

    def __call__(self, element: Element) -> Dict:
        """
        :param element: ``Element`` object, intended to be output of ``fetch()``
        :return: mapping between ``Course`` param names and corresponding values
        """
        data = {}
        for parser in self.parsers:
            vals = parser(element)
            params = parser.params
            data.update(zip(params, vals))
        return data


def generate(*params: str):
    """
    Builds a decorator which wraps functions into ``Parser`` objects.

    :param params: list of param names return values of function are associated with
    :return: decorator which wraps functions into ``Parser`` objects
    """

    def build_parser(parse_func: Callable) -> Parser:
        return Parser(params, parse_func)

    return build_parser


@generate("prereqs", "coreqs", "constraints")
def parse_prereqs(element: Element) -> Tuple[List, List, List]:
    """
    Parses prerequisites from natural language via Lark. For instance,
    ``[{"CS 125", "CS 105"}, {"CS 173"}]`` means "173 and either 105 or 125."

    :param element: ``Element`` object
    :return: list of prereqs, list of coreqs, list of constraints
             (requirements that could not be parsed)
    """
    prereqs, coreqs, constraints = [], [], []
    description = element.find("description")
    if description is not None:
        m = re.search(r"Prerequisite: (.*\.)", description.text)
        if m:
            reqs = re.split("[;.] ", m.group(1))
            for req in reqs:
                data = reduce_tree(PREREQUISITE_PARSER.parse(req))
                prereqs.extend(data.get("prereq", []))
                coreqs.extend(data.get("coreq", []))
                if data.get("other"):
                    constraints.append(req)
    return prereqs, coreqs, constraints


@generate("label")
def parse_label(element: Element) -> str:
    """
    Gets the label of a course (example for CS 173: "discrete structures").

    :param element: Element object
    :return: label text as string
    """
    label = element.find("label")
    if label is not None:
        return label.text


@generate("hours")
def parse_credit_hours(element: Element) -> Optional[int]:
    """
    Gets the number of credit hours of a course.

    :param element: ``Element`` object
    :return: first integer in the creditHours section (if there is one)
    """
    # if something says "3 or 4" hours it's 3 hours...right?
    hours = element.find("creditHours")
    if hours is not None:
        return int(re.search("(\d+)", hours.text).group(1))


@generate("schedule_info")
def parse_schedule_information(element: Element) -> Optional[str]:
    """
    Gets information about course scheduling (this is the courseSectionInfo attribute in the XML).

    :param element: ``Element`` object
    :return: text in courseSectionInformation section (if there is one)
    """
    schedule_info = element.find("courseSectionInformation")
    if schedule_info is not None:
        return schedule_info.text


COMBINATOR = ParserCombinator(parse_prereqs, parse_credit_hours, parse_label,
                              parse_schedule_information)



"""
Utility methods to package UIUC web api data into lxml objects.
"""
import requests
import re
import datetime as dt
from typing import Optional, Iterator
from concurrent.futures import ThreadPoolExecutor
from lxml import etree
from lxml.etree import Element

API_URL = "http://courses.illinois.edu/cisapp/explorer/catalog"
QUARTERS = ("spring", "summer", "fall", "winter")


def fetch(year: Optional[str] = None, quarter: Optional[str] = None, subject: Optional[str] = None,
          number: Optional[str] = None) -> Element:
    """
    Queries official xml api with given parameters.

    :param year: year to get catalog from (defaults to current year)
    :param quarter: one of ``{'fall', 'spring', 'winter', 'summer'}``
                    (any case, defaults to spring)
    :param subject: subject abbreviation
    :param number: 3 digit course number
    :return: appropriate ``lxml.etree.Element`` object
    """
    params = []

    # lots of checks
    for param in (year, quarter, subject, number):
        if param is not None and not isinstance(param, str):
            raise TypeError("non-string data {} passed to fetch".format(param))

    if year is None:
        year = str(dt.datetime.now().year)
    params.append(year)

    if quarter is not None:
        quarter = quarter.lower()
        if quarter not in QUARTERS:
            raise ValueError("quarter must be one of {}".format(", ".join(QUARTERS)))
    else:
        quarter = "spring"
    params.append(quarter)

    if subject is not None:
        params.append(subject)

    if number is not None:
        if subject is None:
            raise ValueError("course number passed without specifying a subject")
        if not re.match(r"\d{3}", number):
            raise ValueError("course number must be 3 digit string")
        params.append(number)

    # parse xml from constructed url
    url = "/".join((API_URL, *params)) + ".xml"
    xml_raw = requests.get(url).content
    try:
        return etree.fromstring(xml_raw)
    except etree.XMLSyntaxError:
        raise ValueError("could not fetch api data for {0} {1}".format(subject, number))


def get_subject_catalog(subject: Optional[str], year: Optional[str] = None,
                        quarter: Optional[str] = None) -> Iterator[str]:
    """
    Gets all the courses for a given subject.

    :param subject: subject abbreviation
    :param year: year to get catalog from (defaults to current year)
    :param quarter: one of ``{'fall', 'spring', 'winter', 'summer'}``
                    (any case, defaults to spring)
    :return: iterator of course names
    """
    catalog = fetch(year=year, quarter=quarter, subject=subject)
    return (
        "{0} {1}".format(subject, course.attrib["id"])
        for course in catalog.find("courses")
    )


def subject_iterator(year: Optional[str] = None, quarter: Optional[str] = None) -> Iterator[str]:
    """
    Gets an iterator of subject names.

    :param year: year to get catalog (defaults to current year)
    :param quarter: one of ``{'fall', 'spring', 'winter', 'summer'}``
                    (any case, defaults to spring)
    :return: generator yielding abbreviated subject names
    """
    catalog = fetch(year=year, quarter=quarter)
    return (e.attrib["id"] for e in catalog.find("subjects"))


def course_iterator(subjects: Optional[Iterator[str]] = None, max_workers: int = 32) -> Iterator[
    str]:
    """
    Gets an iterator of courses, high volume of https requests is streamlined through concurrency.

    :param subjects: iterator of subject names, intended to be output of ``get_subjects()``
                     (defaults to all subjects)
    :param max_workers: max number of worker threads to use
    :return: generator yielding course names: ``('AAS 100', 'AAS 105', 'AAS 120'..)``.
    """
    if subjects is None:
        subjects = subject_names
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        for subject in executor.map(get_subject_catalog, subjects):
            yield from subject


subject_names = subject_iterator()
course_names = course_iterator()



from types import GeneratorType
from lark import Tree


def reduce_tree(tree: Tree):
    """
    Recursively coverts prereq parse tree to dict for further processing.

    :param tree: A subtree of the parse tree
    :return: appropriate object based on root node of subtree passed
    """
    children = tree.children
    rule = tree.data
    if rule == "start":
        return {child.data: reduce_tree(child) for child in children}
    if rule == "course":
        subject, number = children
        subject = "".join(token.value for token in subject.children)
        number = "".join(token.value for token in number.children)
        return "{0} {1}".format(subject, number)
    elif rule == "direct":
        return (frozenset((reduce_tree(child),)) for child in children)
    elif rule == "set":
        return frozenset(reduce_tree(child) for child in children)
    elif rule in ("prereq", "coreq"):
        reqs = []
        for child in children:
            reduced = reduce_tree(child)
            if isinstance(reduced, frozenset):
                reqs.append(reduced)
            elif isinstance(reduced, GeneratorType):
                reqs.extend(reduced)
        return frozenset(reqs)
    elif rule == "other":
        return True
    else:
        raise ValueError("can't reduce tree with unrecognized rule")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="uiuc-api",
    version="0.1.3",
    author="Timothy Zhou",
    description="A python wrapper for UIUC's official REST API for querying course data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/TimothyZhou/uiuc_api",
    packages=setuptools.find_packages(),
    install_requires=[
        'requests', 'lxml', 'pyyaml', 'lark-parser'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    package_data={
        "": ["*.lark"],
    }
)
